/* 
 * pLinguaCore: A JAVA library for Membrane Computing
 *              http://www.p-lingua.org
 *
 * Copyright (C) 2009  Research Group on Natural Computing
 *                     http://www.gcn.us.es
 *                      
 * This file is part of pLinguaCore.
 *
 * pLinguaCore is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * pLinguaCore is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with pLinguaCore.  If not, see <http://www.gnu.org/licenses/>.
 */


package org.gcn.plinguacore.parser.output.binary;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.gcn.plinguacore.parser.output.OutputParser;
import org.gcn.plinguacore.util.MultiSet;
import org.gcn.plinguacore.util.psystem.AlphabetObject;
import org.gcn.plinguacore.util.psystem.Psystem;
import org.gcn.plinguacore.util.psystem.cellLike.CellLikePsystem;
import org.gcn.plinguacore.util.psystem.cellLike.membrane.CellLikeMembrane;
import org.gcn.plinguacore.util.psystem.cellLike.membrane.CellLikeNoSkinMembrane;
import org.gcn.plinguacore.util.psystem.cellLike.membrane.CellLikeSkinMembrane;
import org.gcn.plinguacore.util.psystem.membrane.Membrane;
import org.gcn.plinguacore.util.psystem.rule.Rule;
import org.gcn.plinguacore.util.psystem.rule.checkRule.CheckRule;
import org.gcn.plinguacore.util.psystem.rule.checkRule.specificCheckRule.NoDissolution;
import org.gcn.plinguacore.util.psystem.rule.checkRule.specificCheckRule.NoDivision;
import org.gcn.plinguacore.util.psystem.rule.checkRule.specificCheckRule.NoEvolution;
import org.gcn.plinguacore.util.psystem.rule.checkRule.specificCheckRule.NoLeftExternalMultiSet;

/**
 * This class writes a P system in a binary file
 * 
 *  @author Research Group on Natural Computing (http://www.gcn.us.es)
 * 
 */
public class BinaryCellLikeOutputParser extends OutputParser {

	private Map<String, Integer> alphabet = null;
	Map<String, Integer> labels = null;
	private DataOutputStream stream = null;
	private Psystem psystem = null;
	private Map<Integer, MultiSet<String>> multiSets = null;
	private List<Rule> evolutionRules = null;
	private List<Rule> sendInRules = null;
	private List<Rule> sendOutRules = null;
	private List<Rule> divisionRules = null;
	private List<Rule> dissolutionRules = null;

	private static final CheckRule noEvolutionRule = new NoEvolution(
			);

	private static final CheckRule noDivisionRule = new NoDivision(
			);

	private static final CheckRule noDissolutionRule = new NoDissolution(
			);

	private static final CheckRule noLeftExternalMultiSet = new NoLeftExternalMultiSet(
			);

	public BinaryCellLikeOutputParser() {
		alphabet = new LinkedHashMap<String, Integer>();
		labels = new LinkedHashMap<String, Integer>();
		multiSets = new LinkedHashMap<Integer, MultiSet<String>>();
		evolutionRules = new ArrayList<Rule>();
		sendInRules = new ArrayList<Rule>();
		sendOutRules = new ArrayList<Rule>();
		divisionRules = new ArrayList<Rule>();
		dissolutionRules = new ArrayList<Rule>();
	}

	private void writeHead() throws IOException {
		stream.writeByte(0xAF);
		stream.writeByte(0x12);
		stream.writeByte(0xFA);
		stream.writeByte(0x11);
	}

	private void writeAlphabet() throws IOException {
		alphabet.clear();
		int n = psystem.getAlphabet().size();
		stream.writeChar(n+1);
		alphabet.put("#", 0);
		stream.writeBytes("#");
		stream.writeByte(0);
		Iterator<AlphabetObject> it = psystem.getAlphabet().iterator();
		int i = 1;
		while (it.hasNext()) {
			String obj = it.next().toString();
			if (!alphabet.containsKey(obj)) {
				alphabet.put(obj, i);
				i++;
				stream.writeBytes(obj);
				stream.writeByte(0);
			}
		}

	}

	private void readRules() {
		evolutionRules.clear();
		sendInRules.clear();
		sendOutRules.clear();
		divisionRules.clear();
		dissolutionRules.clear();

		Iterator<Rule> it = psystem.getRules().iterator();

		while (it.hasNext()) {
			Rule r = it.next();

			if (!noEvolutionRule.checkRule(r))
				evolutionRules.add(r);
			else if (!noDivisionRule.checkRule(r))
				divisionRules.add(r);
			else if (!noDissolutionRule.checkRule(r))
				dissolutionRules.add(r);
			else if (noLeftExternalMultiSet.checkRule(r))
				sendOutRules.add(r);
			else
				sendInRules.add(r);

		}
	}
	
	private void writeCharge(byte charge) throws IOException
	{
		if (charge == 0)
			stream.writeByte(0);
		else if (charge > 0)
			stream.writeByte(1);
		else
			stream.writeByte(2);

	}
	
	private void writeObjects(MultiSet<String> objs,boolean withMultiplicity,boolean forced) throws IOException
	{
		if (objs.size()==0 && forced)
		{
			writeObject("#");
			if (withMultiplicity)
				stream.writeChar(1);
		}
		else
		{
			Iterator<String> it=objs.entrySet().iterator();
			while(it.hasNext())
			{
				String obj = it.next();
				writeObject(obj);
				if (withMultiplicity)
					stream.writeChar((int)objs.count(obj));
			}
		}
	}
	
	private void writeObject(String obj) throws IOException
	{
		int id = alphabet.get(obj);
		writeObjectId(id);
	}
	
	private void writeRuleHead(Rule r) throws IOException
	{
		writeLabelId(labels.get(r.getLeftHandRule().getOuterRuleMembrane()
				.getLabel()));
		writeCharge(r.getLeftHandRule().getOuterRuleMembrane().getCharge());
	}

	private void writeSendInRules() throws IOException {
		stream.writeChar(sendInRules.size());
		Iterator<Rule> it = sendInRules.iterator();
		while(it.hasNext())
		{
			Rule r = it.next();
			writeRuleHead(r);
			writeCharge(r.getRightHandRule().getOuterRuleMembrane().getCharge());
			writeObjects(r.getLeftHandRule().getMultiSet(),false,true);
			writeObjects(r.getRightHandRule().getOuterRuleMembrane().getMultiSet(),false,true);
		}
	}
	
	private void writeDissolutionRules() throws IOException {
		stream.writeChar(dissolutionRules.size());
		Iterator<Rule> it = dissolutionRules.iterator();
		while(it.hasNext())
		{
			Rule r = it.next();
			writeRuleHead(r);
			writeObjects(r.getLeftHandRule().getOuterRuleMembrane().getMultiSet(),false,true);
			if (!r.getRightHandRule().getMultiSet().isEmpty())
				writeObjects(r.getRightHandRule().getMultiSet(),false,true);
			else
				writeObjects(r.getRightHandRule().getOuterRuleMembrane().getMultiSet(),false,true);
		}
	}
	
	private void writeDivisionRules() throws IOException {
		stream.writeChar(divisionRules.size());
		Iterator<Rule> it = divisionRules.iterator();
		while(it.hasNext())
		{
			Rule r = it.next();
			writeRuleHead(r);
			writeCharge(r.getRightHandRule().getOuterRuleMembrane().getCharge());
			writeCharge(r.getRightHandRule().getSecondOuterRuleMembrane().getCharge());
			writeObjects(r.getLeftHandRule().getOuterRuleMembrane().getMultiSet(),false,true);
			writeObjects(r.getRightHandRule().getOuterRuleMembrane().getMultiSet(),false,true);
			writeObjects(r.getRightHandRule().getSecondOuterRuleMembrane().getMultiSet(),false,true);
		}
	}
	
	
	private void writeSendOutRules() throws IOException {
		stream.writeChar(sendOutRules.size());
		Iterator<Rule> it = sendOutRules.iterator();
		while(it.hasNext())
		{
			Rule r = it.next();
			writeRuleHead(r);
			writeCharge(r.getRightHandRule().getOuterRuleMembrane().getCharge());
			writeObjects(r.getLeftHandRule().getOuterRuleMembrane().getMultiSet(),false,true);
			writeObjects(r.getRightHandRule().getMultiSet(),false,true);
		}
	}
	
	private void writeEvolutionRules() throws IOException {
		stream.writeChar(evolutionRules.size());
		Iterator<Rule> it = evolutionRules.iterator();
		while (it.hasNext()) {
			Rule r = it.next();
			writeRuleHead(r);
			writeObjects(r.getLeftHandRule().getOuterRuleMembrane().getMultiSet(),false,true);
			stream.writeChar(r.getRightHandRule().getOuterRuleMembrane().getMultiSet().entrySet().size());
			writeObjects(r.getRightHandRule().getOuterRuleMembrane().getMultiSet(),true,false);
		}
	}

	private void writeRules() throws IOException {
		readRules();
		writeEvolutionRules();
		writeSendInRules();
		writeSendOutRules();
		writeDissolutionRules();
		writeDivisionRules();
	}

	private void writeId(int id, int size) throws IOException {
		if (size > 0xFF)
			stream.writeChar(id);
		else
			stream.writeByte(id);
	}

	private void writeMembraneId(int id) throws IOException {
		writeId(id, psystem.getMembranes().size());
	}

	private void writeLabelId(int id) throws IOException {
		writeId(id, labels.size());
	}

	private void writeObjectId(int id) throws IOException {
		writeId(id, alphabet.size());
	}

	private void writeMembrane(CellLikeMembrane membrane) throws IOException {
		if (membrane instanceof CellLikeSkinMembrane)
			writeMembraneId(0);
		else
			writeMembraneId(((CellLikeNoSkinMembrane) membrane)
					.getParentMembrane().getId());
		
		if (psystem.getInitialMultiSets().get(membrane.getLabel())!=null && 
				!psystem.getInitialMultiSets().get(membrane.getLabel()).isEmpty())
		{
			multiSets.put(membrane.getId(),psystem.getInitialMultiSets().get(membrane.getLabel()));
		}
		writeLabelId(labels.get(membrane.getLabel()));
		if (membrane.getCharge() == 0)
			stream.writeByte(0);
		else if (membrane.getCharge() > 0)
			stream.writeByte(1);
		else
			stream.writeByte(2);
		Iterator<CellLikeNoSkinMembrane> it = membrane.getChildMembranes()
				.iterator();
		while (it.hasNext())
			writeMembrane(it.next());

	}

	private void writeMultiSets() throws IOException {
		stream.writeChar(multiSets.size());
		Iterator<Integer> it = multiSets.keySet().iterator();
		while (it.hasNext()) {
			int id = it.next();
			writeMembraneId(id);
			MultiSet<String> ms = multiSets.get(id);
			stream.writeChar(ms.entrySet().size());
			Iterator<String> it1 = ms.entrySet().iterator();
			while (it1.hasNext()) {
				String obj = it1.next();
				long mul = ms.count(obj);
				int objId = alphabet.get(obj);
				writeObjectId(objId);
				stream.writeChar((int)mul);
			}
		}
	}

	private void writeMembranes() throws IOException {
		int n = psystem.getMembranes().size();
		multiSets.clear();
		stream.writeChar(n);
		CellLikeSkinMembrane skin = ((CellLikePsystem) psystem)
				.getSkinMembrane();
		if (skin != null)
			writeMembrane(skin);

	}

	private void writeLabels() throws IOException {
		labels.clear();
		Iterator<? extends Membrane> it = psystem.getMembranes().iterator();
		int i = 0;
		while (it.hasNext()) {
			String label = it.next().getLabel();
			if (!labels.containsKey(label)) {
				labels.put(label, i);
				i++;
			}
		}
		stream.writeChar(labels.size());
		Iterator<String> it1 = labels.keySet().iterator();
		while (it1.hasNext()) {
			stream.writeBytes(it1.next());
			stream.writeByte(0);
		}

	}

	/* (non-Javadoc)
	 * @see org.gcn.plinguacore.parser.output.OutputParser#parse(org.gcn.plinguacore.util.psystem.Psystem, java.io.OutputStream)
	 */
	@Override
	public boolean parse(Psystem psystem, OutputStream stream) {
		if (psystem == null)
			return false;

		if (!psystem.getAbstractPsystemFactory().getModelName().equals(
				"membrane_division"))
			return false;

		try {
			this.psystem = psystem;
			this.stream = new DataOutputStream(stream);
			writeHead();
			writeAlphabet();
			writeLabels();
			writeMembranes();
			writeMultiSets();
			writeRules();
			this.stream.close();

		} catch (IOException e) {

			return false;
		}
		return true;

	}

	/* (non-Javadoc)
	 * @see org.gcn.plinguacore.parser.output.OutputParser#parse(org.gcn.plinguacore.util.psystem.Psystem, java.io.Writer)
	 */
	@Override
	public boolean parse(Psystem psystem, Writer stream) {

		throw new UnsupportedOperationException();
	}

}
