/* 
 * pLinguaCore: A JAVA library for Membrane Computing
 *              http://www.p-lingua.org
 *
 * Copyright (C) 2009  Research Group on Natural Computing
 *                     http://www.gcn.us.es
 *                      
 * This file is part of pLinguaCore.
 *
 * pLinguaCore is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * pLinguaCore is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with pLinguaCore.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.gcn.plinguacore.simulator.cellLike;

import java.util.HashMap;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Map;

import org.gcn.plinguacore.simulator.Simulator;
import org.gcn.plinguacore.util.HashMultiSet;
import org.gcn.plinguacore.util.MultiSet;
import org.gcn.plinguacore.util.Pair;
import org.gcn.plinguacore.util.PlinguaCoreException;
import org.gcn.plinguacore.util.psystem.Configuration;
import org.gcn.plinguacore.util.psystem.Psystem;
import org.gcn.plinguacore.util.psystem.cellLike.CellLikeConfiguration;
import org.gcn.plinguacore.util.psystem.cellLike.CellLikePsystem;
import org.gcn.plinguacore.util.psystem.cellLike.membrane.CellLikeMembrane;
import org.gcn.plinguacore.util.psystem.cellLike.membrane.CellLikeNoSkinMembrane;
import org.gcn.plinguacore.util.psystem.membrane.Membrane;
import org.gcn.plinguacore.util.psystem.rule.InnerRuleMembrane;
import org.gcn.plinguacore.util.psystem.rule.LeftHandRule;
import org.gcn.plinguacore.util.psystem.rule.Rule;
import org.gcn.plinguacore.util.psystem.rule.checkRule.CheckRule;
import org.gcn.plinguacore.util.psystem.rule.checkRule.specificCheckRule.NoEvolution;

/**
 * A simulator for cell-like P systems
 * 
 * @author Research Group on Natural Computing (http://www.gcn.us.es)
 * 
 */
public abstract class CellLikeSimulator extends Simulator {
	private boolean firstTime = true;
	private static final long serialVersionUID = 2539727644482041974L;

	private Map<Integer, Pair<CellLikeMembrane, MultiSet<Rule>>> selectedRules;
	protected static final CheckRule noEvolution = new NoEvolution();

	/**
	 * @param psystem
	 */
	public CellLikeSimulator(Psystem psystem) {
		super(psystem);
		if (!(psystem instanceof CellLikePsystem))
			throw new IllegalArgumentException(
					"psystem argument must be a Cell-Like Psystem");
		currentConfig = psystem.getFirstConfiguration();
		selectedRules = new HashMap<Integer, Pair<CellLikeMembrane, MultiSet<Rule>>>();

	}

	/**
	 * Print short information in the info-channel about current configuration
	 * 
	 * @param first
	 *            It is true if this is the first configuration
	 */
	public void printInfoShort(boolean first) {
		if (!selectedRules.isEmpty() || first) {

			getInfoChannel().println(
					"***********************************************");
			getInfoChannel().println();
			getInfoChannel().println(
					"    CONFIGURATION: " + (currentConfig.getNumber()));
			if (isTimed()) {
				long mem = Runtime.getRuntime().maxMemory()
						- Runtime.getRuntime().freeMemory();
				mem = mem / 1024;
				getInfoChannel().println("    TIME: " + getTime() + " s.");
				getInfoChannel().println("    MEMORY: " + mem + " Kb");
			}
			getInfoChannel().println();

			CellLikeConfiguration c = (CellLikeConfiguration) currentConfig;
			printInfo(c.getSkinMembrane());
			if (!currentConfig.getEnvironment().isEmpty()) {
				getInfoChannel().println(
						"    ENVIRONMENT: " + currentConfig.getEnvironment());
				getInfoChannel().println();
			}
		} else {
			Iterator<? extends Membrane> it = currentConfig.getMembranes()
					.iterator();
			while (it.hasNext())
				printInfo((CellLikeMembrane) it.next());
			if (!currentConfig.getEnvironment().isEmpty()) {
				getInfoChannel().println(
						"    ENVIRONMENT: " + currentConfig.getEnvironment());
				getInfoChannel().println();
			}
			getInfoChannel()
					.println(
							"Halting configuration (No rule can be selected to be executed in the next step)");
		}

	}

	/**
	 * Print large information in the info-channel about current configuration
	 * 
	 * @param first
	 *            It is true if this is the first configuration
	 */
	public void printInfo(boolean first) {
		if (!selectedRules.isEmpty() || first) {
			if (!first) {
				getInfoChannel().println(
						"-----------------------------------------------");
				getInfoChannel().println();
				getInfoChannel().println(
						"    STEP: " + currentConfig.getNumber());

			}
			Iterator<Pair<CellLikeMembrane, MultiSet<Rule>>> it1 = selectedRules
					.values().iterator();

			while (it1.hasNext()) {
				Pair<CellLikeMembrane, MultiSet<Rule>> pair = it1.next();
				MultiSet<Rule> rules = pair.getSecond();
				Iterator<Rule> it = rules.entrySet().iterator();
				if (it.hasNext()) {
					getInfoChannel().println();
					getInfoChannel().println(
							"    Rules selected for "
									+ getHead(pair.getFirst()));

				}
				while (it.hasNext()) {
					Rule r = it.next();
					getInfoChannel().println(
							"    " + rules.count(r) + " * " + r.toString());
				}
			}
			getInfoChannel().println();
			getInfoChannel().println(
					"***********************************************");
			getInfoChannel().println();
			getInfoChannel().println(
					"    CONFIGURATION: " + (currentConfig.getNumber()));
			if (isTimed()) {
				getInfoChannel().println("    TIME: " + getTime() + " s.");
				getInfoChannel().println(
						"    MEMORY USED: "
								+ Runtime.getRuntime().totalMemory() / 1024);
				getInfoChannel().println(
						"    FREE MEMORY: " + Runtime.getRuntime().freeMemory()
								/ 1024);
				getInfoChannel().println(
						"    TOTAL MEMORY: " + Runtime.getRuntime().maxMemory()
								/ 1024);
			}
			getInfoChannel().println();

			Iterator<? extends Membrane> it = currentConfig.getMembranes()
					.iterator();
			while (it.hasNext())
				printInfo((CellLikeMembrane) it.next());
			if (!currentConfig.getEnvironment().isEmpty()) {
				getInfoChannel().println(
						"    ENVIRONMENT: " + currentConfig.getEnvironment());
				getInfoChannel().println();
			}
		} else {
			getInfoChannel()
					.println(
							"Halting configuration (No rule can be selected to be executed in the next step)");
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.gcn.plinguacore.simulator.Simulator#reset()
	 */
	@Override
	public void reset() {
		// TODO Auto-generated method stub
		super.reset();

		printInfo(true);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.gcn.plinguacore.simulator.Simulator#specificStep()
	 */
	@Override
	protected boolean specificStep() throws PlinguaCoreException {

		if (firstTime) {
			if (getVerbosity() > 0)
				printInfo(true);
			else
				printInfoShort(true);
			firstTime = false;
		}

		microStepInit();
		microStepSelectRules();

		if (!selectedRules.isEmpty()) {
			microStepExecuteRules();
			currentConfig.setNumber(currentConfig.getNumber() + 1);

		}

		if (getVerbosity() > 0)
			printInfo(false);
		else
			printInfoShort(false);
	
		return !selectedRules.isEmpty();
	}

	private void microStepExecuteRules() {

		Iterator<Pair<CellLikeMembrane, MultiSet<Rule>>> it = selectedRules
				.values().iterator();
		MultiSet<Rule> ms1 = new HashMultiSet<Rule>();
		while (it.hasNext()) {
			Pair<CellLikeMembrane, MultiSet<Rule>> p = it.next();
			MultiSet<Rule> ms = p.getSecond();
			CellLikeMembrane m = p.getFirst();
			ms1.clear();
			Iterator<Rule> it1 = ms.entrySet().iterator();

			while (it1.hasNext()) {
				Rule r = it1.next();
				if (noEvolution.checkRule(r))
					ms1.add(r, ms.count(r));
				else
					r.execute(m, currentConfig.getEnvironment(), ms
							.count(r));
			}
			it1 = ms1.entrySet().iterator();
			while (it1.hasNext()) {
				Rule r = it1.next();
				r.execute(m, currentConfig.getEnvironment(), ms1
						.count(r));
			}

		}

	}

	protected void selectRule(Rule r, CellLikeMembrane m, long count) {
		Pair<CellLikeMembrane, MultiSet<Rule>> p;
		if (selectedRules.containsKey(m.getId()))
			p = selectedRules.get(m.getId());
		else {
			p = new Pair<CellLikeMembrane, MultiSet<Rule>>(m,
					new HashMultiSet<Rule>());
			selectedRules.put(m.getId(), p);
		}
		p.getSecond().add(r, count);
	}

	private String getHead(CellLikeMembrane m) {
		String str = "";
		if (m.isSkinMembrane())
			str += "SKIN MEMBRANE ID: ";
		else
			str += "MEMBRANE ID: ";
		str += m.getId() + ", Label: " + m.getLabelObj() + ", Charge: "
				+ Membrane.getChargeSymbol(m.getCharge());
		return str;
	}

	private void printInfo(CellLikeMembrane m) {

		getInfoChannel().println("    " + getHead(m));
		getInfoChannel().println("    Multiset: " + m.getMultiSet());
		if (!m.getChildMembranes().isEmpty())
			getInfoChannel().println(
					"    Internal membranes count: "
							+ m.getChildMembranes().size());
		if (!m.isSkinMembrane())
			getInfoChannel().println(
					"    Parent membrane ID: "
							+ ((CellLikeNoSkinMembrane) m).getParentMembrane()
									.getId());

		getInfoChannel().println();

	}

	protected void microStepInit() {

		selectedRules.clear();
		initDate();
	}

	protected void microStepSelectRules() {

		Iterator<? extends Membrane> it = ((Configuration)currentConfig.clone()).getMembranes()
				.iterator();

		Iterator<? extends Membrane> it1 = currentConfig.getMembranes().iterator();
		
		while (it.hasNext()) {
			CellLikeMembrane tempMembrane = (CellLikeMembrane) it.next();
			CellLikeMembrane m = (CellLikeMembrane)it1.next();
			microStepSelectRules(m,tempMembrane);

		}

	}

	/**
	 * Select rules for a specific membrane
	 * 
	 * @param m
	 *            A membrane
	 */
	protected abstract void microStepSelectRules(CellLikeMembrane membrane,CellLikeMembrane tempMembrane);

	protected static void removeLeftHandRuleObjects(CellLikeMembrane m, Rule r,
			long count) {
		LeftHandRule lhr = r.getLeftHandRule();
		if (!m.isSkinMembrane()) {
			CellLikeNoSkinMembrane noSkin = (CellLikeNoSkinMembrane) m;
			noSkin.getParentMembrane().getMultiSet().subtraction(
					lhr.getMultiSet(), count);
		}
		m.getMultiSet().subtraction(lhr.getOuterRuleMembrane().getMultiSet(),
				count);
		ListIterator<InnerRuleMembrane> it = lhr.getOuterRuleMembrane()
				.getInnerRuleMembranes().listIterator();

		while (it.hasNext()) {
			InnerRuleMembrane im = it.next();
			Iterator<CellLikeNoSkinMembrane> it1 = m.getChildMembranes()
					.iterator();
			while (it1.hasNext()) {
				CellLikeNoSkinMembrane m1 = it1.next();
				if (m1.getCharge() == im.getCharge()
						&& m1.getLabel().equals(im.getLabel())) {
					m1.getMultiSet().subtraction(im.getMultiSet(), count);
					break;
				}
			}
		}
	}

	@Override
	public boolean isFinished() {

		boolean finished = true;
		Iterator<? extends Membrane> it = currentConfig.getMembranes()
				.iterator();

		while (it.hasNext() && finished) {
			CellLikeMembrane m = (CellLikeMembrane) it.next();
			Iterator<Rule> itRules = psystem.getRules().iterator(m.getLabel(),
					m.getCharge());
			while (itRules.hasNext() && finished) {
				if (itRules.next().countExecutions(m) > 0)
					finished = false;
			}
		}

		return finished;
	}

}
