/* 
 * pLinguaCore: A JAVA library for Membrane Computing
 *              http://www.p-lingua.org
 *
 * Copyright (C) 2009  Research Group on Natural Computing
 *                     http://www.gcn.us.es
 *                      
 * This file is part of pLinguaCore.
 *
 * pLinguaCore is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * pLinguaCore is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with pLinguaCore.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.gcn.plinguacore.util.psystem.rule;

import java.io.Serializable;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;

import org.gcn.plinguacore.util.MultiSet;
import org.gcn.plinguacore.util.psystem.membrane.ChangeableMembrane;


/**
 * This class represents all kind of rules, intended to be extended to fulfill
 * the requirements of each P-System group
 * 
 *  @author Research Group on Natural Computing (http://www.gcn.us.es)
 * 
 */
public abstract class Rule implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = -2223028108798909376L;



	protected InnerRuleMembrane[] handCorrespondence;


	private boolean dissolves;
	private boolean empty;
	private LeftHandRule leftHandRule;
	private RightHandRule rightHandRule;
	

	/**
	 * Creates a new Rule instance. As Rule class is abstract, this constructor
	 * is intended to be used by Rule subclasses
	 * 
	 * @param dissolves
	 *            a boolean parameter which reflects if the rule will dissolve
	 *            the membrane
	 * @param leftHandRule
	 *            the left hand of the rule
	 * @param rightHandRule
	 *            the right hand of the rule
	 */
	public Rule(boolean dissolves, LeftHandRule leftHandRule,
			RightHandRule rightHandRule) {
		super();
		if (leftHandRule == null)
			throw new NullPointerException(
					"LeftHandRule parameter in Rule constructor shouldn't be null");
		if (rightHandRule == null)
			throw new NullPointerException(
					"RightHandRule parameter in Rule constructor shouldn't be null");
		this.dissolves = dissolves;
		this.leftHandRule = leftHandRule;
		this.rightHandRule = rightHandRule;
		checkState();
		checkEmpty();
	}

	/**
	 * Gets all different objects within the rule
	 * 
	 * @return a set containing all different objects within the rule
	 * */
	public Set<String> getObjects() {
		Set<String> set = leftHandRule.getObjects();
		set.addAll(rightHandRule.getObjects());
		return set;

	}
	
	


	/**
	 * Gets a boolean reporting if there's not any object in the left side of
	 * the rule
	 * 
	 * @return if there's not any object in the left side of the rule
	 */
	public boolean isEmpty() {
		return empty;
	}

	private void checkEmpty() {
		/*
		 * Now we find out if the left hand rule is empty (devoid of objects in
		 * any contained multiset)
		 */
		/*
		 * It won't be empty if there's any object in the multiset contained
		 * into the left hand rule outer membrane or if there's any object in
		 * the outer multiset
		 */
		empty = (!leftHandRule.getMultiSet().isEmpty())
				|| (!leftHandRule.getOuterRuleMembrane().getMultiSet()
						.isEmpty());

		if (!empty)
			return;
		/* Neither it will be empty if there's any object in any inner membrane */
		Iterator<InnerRuleMembrane> it = leftHandRule.getOuterRuleMembrane()
				.getInnerRuleMembranes().iterator();
		while (it.hasNext() && empty)
			empty = it.next().getMultiSet().isEmpty();

	}

	private void checkState() {

		/*
		 * There are several conditions a rule should comply with: a rule can't
		 * dissolve a membrane and create a new one
		 */
		if (dissolves && rightHandRule.getSecondOuterRuleMembrane() != null)
			throw new IllegalArgumentException(
					"A rule cannot perform a division and a dissolution at once");
		/* Rules outer membranes (on both hands) should be equally labeled */
		if (!leftHandRule.getOuterRuleMembrane().getLabel().equals(
				rightHandRule.getOuterRuleMembrane().getLabel()))
			throw new IllegalArgumentException(
					"Membrane labels on both sides should match");

		/*
		 * If a rule divides a membrane, both labels in the outer membranes
		 * should be equal
		 */
		if (rightHandRule.getSecondOuterRuleMembrane() != null) {
			if (!leftHandRule.getOuterRuleMembrane().getLabel().equals(
					rightHandRule.getSecondOuterRuleMembrane().getLabel()))
				throw new IllegalArgumentException(
						"Membrane labels on both sides should match");
		}
		
	

		setHandCorrespondence();

	}

	/**
	 * Gets if a certain InnerRuleMembrane instance has any other instance which
	 * corresponds to this
	 * 
	 * @param irm
	 *            the InneRuleMembrane instance to be checked for other to
	 *            correspond
	 * @return if a certain InnerRuleMembrane instance has any other instance
	 *         which corresponds to this
	 */
	public boolean hasCorrespondence(InnerRuleMembrane irm) {
		return handCorrespondence[leftHandRule.getOuterRuleMembrane()
				.getInnerRuleMembranes().indexOf(irm)] != null;
	}

	private void setHandCorrespondence() {
		/*
		 * In this method we need to match any inner membranes in the left hand
		 * and any inner membranes in the right hand, and detect those which
		 * have been created
		 */
		List<InnerRuleMembrane> lirm = leftHandRule.getOuterRuleMembrane()
				.getInnerRuleMembranes();
		int size = lirm.size();
		handCorrespondence = new InnerRuleMembrane[size];
		/* At first, there's no correspondence between hands */
		for (int i = 0; i < size; i++)
			handCorrespondence[i] = null;
		ListIterator<InnerRuleMembrane> lirmIterator = lirm.listIterator();
		int leftHandCounter = 0;
		/*
		 * Now we go through each inner rule membrane on the right hand for each
		 * inner rule membrane on the right hand
		 */
		List<InnerRuleMembrane> rirm = rightHandRule.getOuterRuleMembrane()
				.getInnerRuleMembranes();
		while (lirmIterator.hasNext()) {
			InnerRuleMembrane leftirm = lirmIterator.next();
			ListIterator<InnerRuleMembrane> rirmIterator = rirm.listIterator();
			while (rirmIterator.hasNext()) {
				InnerRuleMembrane rightirm = rirmIterator.next();
				if (rightirm.getLabel().equals(leftirm.getLabel())) {
					/*
					 * In case we find two inner membranes with the same label
					 * (each one on a different hand), we assume they correspond
					 * to each other
					 */
					handCorrespondence[leftHandCounter] = rightirm;
					break;
				}
			}
			leftHandCounter++;

		}

	}

	/**
	 * Gets if the rule dissolves the membrane which is applied to
	 * 
	 * @return a boolean representing if the rule dissolves the membrane which
	 *         is applied to
	 */
	public boolean dissolves() {
		return dissolves;
	}

	/**
	 * Gets the left hand rule of the current rule
	 * 
	 * @return the left hand rule of the current rule
	 */
	public LeftHandRule getLeftHandRule() {
		return leftHandRule;
	}

	/**
	 * Gets the right hand rule of the current rule
	 * 
	 * @return the right hand rule of the current rule
	 */
	public RightHandRule getRightHandRule() {
		return rightHandRule;
	}

	/**
	 * Gets if the rule can be applied to the given membrane
	 * 
	 * @param membrane
	 *            the ChangeableMembrane instance which the rule could be
	 *            applied to
	 * @return if the rule can be applied to the given ChangeableMembrane
	 *         instance
	 */
	public boolean isExecutable(ChangeableMembrane membrane) {
		return countExecutions(membrane) > 0;
	}

	/**
	 * Counts the number of possible executions of the current rule on a
	 * membrane
	 * 
	 * @param membrane
	 *            the membrane where the executions will be counted
	 * @return the number of possible executions on the ChangeableMembrane
	 *         instance given
	 */
	public abstract long countExecutions(ChangeableMembrane membrane);

	/**
	 * Executes the rule on a membrane and an outer multiset and returns if the
	 * execution could be committed
	 * 
	 * @param membrane
	 *            the membrane which the rule will be applied to
	 * @param environment
	 *            the multiset out of the membrane
	 * @return if the execution could be committed
	 */
	public boolean execute(ChangeableMembrane membrane,
			MultiSet<String> environment) {

		return execute(membrane, environment, countExecutions(membrane));
	}
	/**
	 * Applies the right hand of the rule to the membrane passed as argument, by updating all objects which the right hand consist of, updating the membrane children and, if specified, dissolving or dividing the rule 
	 * @param membrane the membrane which the right hand rule will be applied to
	 * @param environment the environment which should be expanded by adding the right hand objects in case the rule is applied to the skin membrane
	 * @param executions the number of times the right hand rule is applied to the membrane
	 */
	protected abstract void executeRightHand(ChangeableMembrane membrane,
			MultiSet<String> environment, long executions);

	protected abstract boolean executeSafe(ChangeableMembrane membrane,
			MultiSet<String> environment, long executions); 


	/**
	 * Executes the rule on a membrane and an outer multiset a number of times
	 * indicated by executions and returns if the execution could be committed
	 * 
	 * @param membrane
	 *            the membrane which the rule will be applied to
	 * @param environment
	 *            the multiset out of the membrane
	 * @param executions
	 *            the number of times to be executed on the membrane passed
	 * @return if the execution could be committed
	 */
	public boolean execute(ChangeableMembrane membrane,
			MultiSet<String> environment, long executions) {
		if (executions > countExecutions(membrane))
		{
			System.out.println(executions+" "+countExecutions(membrane)+" "+membrane+" "+toString());
			return false;
		}
		/*
		 * We only execute the rule if we have made sure it can be executed the
		 * number of executions passed as parameter
		 */
		return executeSafe(membrane, environment, executions);
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
	
			return leftHandRule.toString() + " --> " + rightHandRule.toString();
	
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result + (dissolves ? 1231 : 1237);
		result = PRIME * result
				+ ((leftHandRule == null) ? 0 : leftHandRule.hashCode());
		result = PRIME * result
				+ ((rightHandRule == null) ? 0 : rightHandRule.hashCode());
		return result;
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {

		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final Rule other = (Rule) obj;
		if (dissolves != other.dissolves)
			return false;
		if (leftHandRule == null) {
			if (other.leftHandRule != null)
				return false;
		} else if (!leftHandRule.equals(other.leftHandRule))
			return false;
		if (rightHandRule == null) {
			if (other.rightHandRule != null)
				return false;
		} else if (!rightHandRule.equals(other.rightHandRule))
			return false;
		return true;
	}

	protected void subtractMultiSet(MultiSet<String> ruleMultiSet, MultiSet<String> membraneMultiSet, long executions) {
		// TODO Auto-generated method stub
		membraneMultiSet.subtraction(ruleMultiSet, executions);
	
	}

}
