/* 
 * pLinguaCore: A JAVA library for Membrane Computing
 *              http://www.p-lingua.org
 *
 * Copyright (C) 2009  Research Group on Natural Computing
 *                     http://www.gcn.us.es
 *                      
 * This file is part of pLinguaCore.
 *
 * pLinguaCore is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * pLinguaCore is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with pLinguaCore.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.gcn.plinguacore.util.psystem.rule;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.LinkedHashSet;
import java.util.Collection;
import java.util.Iterator;
import java.util.SortedSet;

import java.util.NoSuchElementException;

import org.gcn.plinguacore.util.Pair;
import org.gcn.plinguacore.util.RandomArrayList;
import org.gcn.plinguacore.util.psystem.AlphabetObject;
import org.gcn.plinguacore.util.psystem.membrane.Membrane;
import org.gcn.plinguacore.util.psystem.rule.checkRule.CheckRule;

/**
 * A Set for rules
 *  @author Research Group on Natural Computing (http://www.gcn.us.es)
 *
 */
public class RulesSet implements Set<Rule>, Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 244011999390422931L;
	private CheckRule checkRule;

	private static final Iterator<Rule> emptyIterator = (new LinkedHashSet<Rule>())
			.iterator();
	private Map<String, Set<Rule>[]> rules = null;


	private int size = 0;

	private SortedSet<AlphabetObject> alphabet = null;
/**
 * Constructs an empty RulesSet
 */
	public RulesSet() {
		super();
		rules = new HashMap<String, Set<Rule>[]>();
	}

	/**
	 * Constructs a RulesSet with initial rules 
	 *  @param rules The initial rules
	 */
	public RulesSet(Collection<? extends Rule> rules) {
		super();
		if (rules instanceof RulesSet)
			this.rules = new HashMap<String, Set<Rule>[]>(
					((RulesSet) rules).rules);
		else {
			this.rules = new HashMap<String, Set<Rule>[]>();
			addAll(rules);
		}

	}
/**
 * Constructs a RulesSet and update the alphabet passed as argument each time a new rule is added to the set
 * @param alphabet The alphabet for this RulesSet
 */
	public RulesSet(SortedSet<AlphabetObject> alphabet) {
		this();
		this.alphabet = alphabet;
	}
	/**
	 * Returns true if the rule passed as argument complies the model conditions
	 * @param r The rule to be checked
	 * @return true if the rule complies the model conditions
	 */
	public boolean checkRule(Rule r) {
		if (checkRule == null)
			return true;
		return checkRule.checkRule(r);
	}
	
	/**
	 * Returns true if all the rules within the RuleSet complies the model conditions
	 * @return true if all the rules within the RuleSet complies the model conditions
	 */
	 

	public boolean checkRules() {
		if (checkRule == null)
			return true;
		Iterator<Rule> itRule = iterator();
		while (itRule.hasNext()) {
			if (!checkRule.checkRule(itRule.next()))
				return false;
		}
		return true;
	}

	/**
	 * Returns the CheckRule object for this RulesSet
	 * @return The CheckRule associated to this RulesSet
	 */
	public CheckRule getCheckRule() {
		return checkRule;
	}
	/**
	 * Set the CheckRule object for this RulesSet 
	 * @param cr The CheckRule to be associated
	 */
	public void setCheckRule(CheckRule cr) {
		if (cr == null)
			throw new NullPointerException(
					"CheckRule parameter shouldn't be null");
		this.checkRule = cr;
	}
	
	/**
	 * Returns an Iterator<Rule> for rules with a specific label, charge and environmentID
	 * @param labelID
	 * @param environmentID
	 * @param charge
	 * @return An Iterator<Rule> object for rules with label, charge and environment specified as arguments.
	 * Tambien se incluyen las reglas con la misma carga, etiqueta e identificador de entorno vacio.
	 */
	public Iterator<Rule>iterator(String labelID,String environmentID,int charge,boolean random)
	{
		if (environmentID==null)
			return iterator(labelID,charge);
		Set<Rule>[] set = rules.get(labelID);
		if (set == null)
			return emptyIterator;
		int index = toIndex(charge);
		List<Rule>rules=new ArrayList<Rule>();
		Iterator<Rule>it = set[index].iterator();
		while(it.hasNext())
		{
			Rule r = it.next();
			String envID=r.getLeftHandRule().getOuterRuleMembrane().getLabelObj().getEnvironmentID();
			if (envID.equals("") || envID.equals(environmentID))
				rules.add(r);
		}
		if (random)
			rules = new RandomArrayList<Rule>(rules);
			
		
		return Collections.unmodifiableList(rules).iterator();
	}
	
	/**
	 * Returns an iterator<Rule> for rules with a specific label and charge.
	 * @param labelID
	 * @param charge
	 * @return An iterator<Rule> object for rules with label and charge passed as arguments.
	 */
	public Iterator<Rule> iterator(String labelID, int charge) {
		Set<Rule>[] set = rules.get(labelID);
		if (set == null)
			return emptyIterator;
		int index = toIndex(charge);
		return Collections.unmodifiableSet(set[index]).iterator();
	}

	private int toIndex(int charge) {
		int index;
		if (charge < 0)
			index = 0;
		else if (charge == 0)
			index = 1;
		else
			index = 2;
		return index;
	}

	@Override
	public boolean add(Rule rule) {
	
		String label = rule.getLeftHandRule().getOuterRuleMembrane().getLabel();
		int index = toIndex(rule.getLeftHandRule().getOuterRuleMembrane()
				.getCharge());
	
		Set<Rule>[] list = rules.get(label);
		if (list == null) {
			Set<Rule>[] rulesByCharge = new Set[3];
			for (int i = 0; i < 3; i++)
				rulesByCharge[i] = new LinkedHashSet<Rule>();
			rules.put(label, rulesByCharge);
			list = rulesByCharge;
		}
	
		boolean b = list[index].add(rule);
		if (b) {
		
			size++;
			if (alphabet != null) {
				Iterator<String> objs = rule.getObjects().iterator();
				while (objs.hasNext())
					alphabet.add(new AlphabetObject(objs.next()));
			}
		}
	
		return b;

	}

	@Override
	public String toString() {

		String str = "";
		Iterator<String> it = rules.keySet().iterator();

		while (it.hasNext()) {
			String label = it.next();
			Set<Rule>[] s = rules.get(label);
			for (int i = 0; i < 3; i++) {
				if (s[i].size() > 0)
					str += "\nLabel: " + label + ", Charge: "
							+ Membrane.getChargeSymbol((byte) (i - 1)) + "\r\n";
				Iterator<Rule> it1 = s[i].iterator();
				while (it1.hasNext())
					str += it1.next() + "\r\n";
			}
		}

		return str;
	}

	@Override
	public boolean addAll(Collection<? extends Rule> arg0) {
		Iterator<? extends Rule> it = arg0.iterator();
		boolean b = false;
		while (it.hasNext())
			b = add(it.next()) || b;
		return b;
	}

	@Override
	public void clear() {
		rules.clear();
		
		size = 0;
	}

	@Override
	public boolean contains(Object arg0) {
		
		if (!(arg0 instanceof Rule))
			return false;
		Rule r = (Rule) arg0;
		Set<Rule>[] s = rules.get(r.getLeftHandRule().getOuterRuleMembrane()
				.getLabel());
		if (s == null)
			return false;
		int index = toIndex(r.getLeftHandRule().getOuterRuleMembrane()
				.getCharge());
		
		return s[index].contains(arg0);

	}

	@Override
	public boolean containsAll(Collection<?> arg0) {
		Iterator<?> it = arg0.iterator();
		boolean b = true;
		while (it.hasNext() && b)
			b = b && contains(it.next());
		return b;

	}

	@Override
	public boolean isEmpty() {
		return rules.isEmpty();
	}

	@Override
	public Iterator<Rule> iterator() {
		return new RulesSetIterator();
	}

	@Override
	public boolean remove(Object arg0) {
		if (!(arg0 instanceof Rule))
			return false;
		Rule r = (Rule) arg0;
		Set<Rule>[] s = rules.get(r.getLeftHandRule().getOuterRuleMembrane()
				.getLabel());
		if (s == null)
			return false;
		int index = toIndex(r.getLeftHandRule().getOuterRuleMembrane()
				.getCharge());
		boolean b = s[index].remove(arg0);
		if (b) {
			size--;
			
		}

		return b;
	}

	@Override
	public boolean removeAll(Collection<?> arg0) {
		Iterator<?> it = arg0.iterator();
		boolean b = false;
		while (it.hasNext())
			/* QUERY: the former code didn't care about lazy evaluation */
			b = remove(it.next()) || b;
		return b;
	}

	@Override
	public boolean retainAll(Collection<?> arg0) {
		Iterator<Rule> it = iterator();
		boolean b = false;
		while (it.hasNext()) {
			Rule r = it.next();
			if (!(arg0.contains(r))) {
				it.remove();
				b = true;
			}
		}
		return b;
	}

	@Override
	public int size() {
		return size;
	}

	@Override
	public Object[] toArray() {
		Object o[] = new Object[size];
		Iterator<Rule> it = iterator();
		int i = 0;
		while (it.hasNext()) {
			o[i] = it.next();
			i++;
		}
		return o;
	}

	@Override
	public <T> T[] toArray(T[] arg0) {
		T[] aux;
		if (arg0 == null)
			throw new NullPointerException();
		if (arg0.length < size)
			aux = (T[]) new Object[size];
		else if (arg0.length == size)
			aux = arg0;
		else {
			aux = arg0;
			aux[size] = null;
		}

		Iterator it = iterator();
		int i = 0;
		while (it.hasNext()) {

			Object obj = it.next();
			if (aux[i].getClass().isAssignableFrom(obj.getClass()))
				aux[i] = (T) obj;
			else
				throw new ArrayStoreException();
			i++;
		}
		return aux;
	}

	private class RulesSetIterator implements Iterator<Rule> {
		private int index = 0;
		private Iterator<String> it;
		private String currentLabel = null;
		private Iterator<Rule> it1 = null;
		private int state = 0;
		private int count = 0;

		public RulesSetIterator() {
			it = rules.keySet().iterator();
		}

		public boolean hasNext() {
			return count < size;
		}

		public Rule next() {
			switch (state) {
			case 0:
				currentLabel = it.next();
				state = 1;
				return next();

			case 1:
				it1 = rules.get(currentLabel)[index].iterator();
				state = 2;
				return next();

			case 2:
				try {
					Rule r = it1.next();
					count++;
					return r;
				} catch (NoSuchElementException ex) {
					if (index < 2) {
						index++;
						state = 1;
						return next();
					} else {
						index = 0;
						state = 0;
						return next();
					}
				}

			default:
				throw new NoSuchElementException();

			}
		}

		public void remove() {
			if (it1 == null)
				throw new IllegalStateException();
			it1.remove();
		}

	}

}
